@@ -1,7 +1,7 @@
 package Bencher;
 
-our $DATE = '2015-12-17'; # DATE
-our $VERSION = '0.18'; # VERSION
+our $DATE = '2015-12-23'; # DATE
+our $VERSION = '0.19'; # VERSION
 
 use 5.010001;
 use strict;
@@ -419,15 +419,15 @@ sub _gen_items {
     if ($datasets && @$datasets) {
         push @permute, "dataset", [map {$_->{seq}} @$datasets];
     }
+
     $log->debugf("permute: %s", \@permute);
 
     my $iter = Permute::Named::Iter::permute_named_iter(@permute);
-    my $i = -1;
+    my $item_seq = -1;
     my $items = [];
   ITER:
     while (my $h = $iter->()) {
         $log->tracef("iter returns: %s", $h);
-        $i++;
         my $item_name;
 
         my $p = _find_record_by_seq($participants, $h->{participant});
@@ -492,83 +492,126 @@ sub _gen_items {
             }
         }
 
-        {
-            # convert participant's & dataset index to name temporarily, for
-            # nicer item name
-            my %h = %$h;
-            $h{participant} = $p->{name};
-            $h{dataset} = $ds->{name} if $ds;
-            my @k = keys %h;
-            if (@k == 1) {
-                $item_name = $h{$k[0]};
-            } else {
-                $item_name = dmp(\%h);
-            }
-            #$log->tracef("Set item name to: %s", $item_name);
-        }
-
-        my $code;
-        if ($p->{type} eq 'command') {
-            my @cmd;
-            my $shell;
-            if (ref($p->{cmdline}) eq 'ARRAY') {
-                @cmd = @{ $p->{cmdline} };
-                $shell = 0;
-            } else {
-                @cmd = ($p->{cmdline});
-                $shell = 1;
-            }
-            $code = sub {
-                if ($shell) {
-                    system $cmd[0];
-                } else {
-                    system {$cmd[0]} @cmd;
-                }
-                die "Command failed (child error=$?, os error=$!)\n"
-                    if $?;
-            };
-        } elsif ($p->{type} eq 'perl_code') {
-            if ($p->{code}) {
-                if ($ds) {
-                    if ($ds->{argv}) {
-                        $code = sub { $p->{code}->(@{$ds->{argv}}) };
-                    } elsif ($ds->{args}) {
-                        $code = sub { $p->{code}->(%{$ds->{args}}) };
-                    } else {
-                        return [400, "Participant #$p->{seq}, dataset #$h->{dataset}: No argv/args supplied for code"];
-                    }
-                } else {
-                    $code = $p->{code};
-                }
-            } elsif (my $template = $p->{code_template} || $p->{fcall_template}) {
-                my $template_vars;
-                if ($ds->{args}) {
-                    $template_vars = $ds->{args};
-                } elsif ($ds->{argv}) {
-                    $template_vars = { map {$_=>$ds->{argv}[$_]}
-                                           0..$#{ $ds->{argv} } };
-                } else {
-                    #warn "Item #$i: participant specifies code_template/fcall_template but there is no args/argv in the dataset #$h->{dataset}\n";
-                }
-
-                if ($template_vars) {
-                    $template =~ s/\<(\w+)\>/dmp($template_vars->{$1})/eg;
-                }
-                my $code_str = "sub { $template }";
-                $log->debugf("Item #%d: code=%s", $i, $code_str);
-                $code = eval $code_str;
-                return [400, "Item #$i: code compile error: $@ (code: $code_str)"] if $@;
+        my $iter_args;
+        if ($ds && $ds->{args} &&
+                (my @multi_keys = grep {/\@\z/} keys %{$ds->{args}})) {
+            # we need to permute arguments also
+            my @permute_args;
+            for (@multi_keys) {
+                s/\@\z//;
+                push @permute_args, $_ => $ds->{args}{"$_\@"};
             }
+            $iter_args = Permute::Named::Iter::permute_named_iter(
+                @permute_args);
+            $log->debugf("permute args: %s", \@permute_args);
         } else {
-            return [400, "Unknown participant type '$p->{type}'"];
+            # create an iterator that returns just a single item: {}
+            # require Array::Iter; $iter_args = Array::Iter::list_iter({});
+            $iter_args = do {
+                my $ary = [{}];
+                my $i = 0;
+                sub {
+                    if ($i < @$ary) {
+                        return $ary->[$i++];
+                    } else {
+                        return undef;
+                    }
+                };
+            };
         }
 
-        push @$items, {
-            seq  => $i,
-            name => $item_name,
-            code => $code,
-            _permute => $h,
-        };
+      ITER_ARGS:
+        while (my $h_args = $iter_args->()) {
+            $item_seq++;
+            my $code;
+            if ($p->{type} eq 'command') {
+                my @cmd;
+                my $shell;
+                if (ref($p->{cmdline}) eq 'ARRAY') {
+                    @cmd = @{ $p->{cmdline} };
+                    $shell = 0;
+                } else {
+                    @cmd = ($p->{cmdline});
+                    $shell = 1;
+                }
+                $code = sub {
+                    if ($shell) {
+                        system $cmd[0];
+                    } else {
+                        system {$cmd[0]} @cmd;
+                    }
+                    die "Command failed (child error=$?, os error=$!)\n"
+                        if $?;
+                };
+            } elsif ($p->{type} eq 'perl_code') {
+                my $args;
+                if ($ds->{args}) {
+                    $args = { %{$ds->{args}} };
+                    delete $args->{$_} for (grep {/\@\z/} keys %$args);
+                    $args->{$_} = $h_args->{$_} for keys %$h_args;
+                }
+                if ($p->{code}) {
+                    if ($ds) {
+                        if ($ds->{argv}) {
+                            $code = sub { $p->{code}->(@{$ds->{argv}}) };
+                        } elsif ($ds->{args}) {
+                            $code = sub { $p->{code}->(%$args) };
+                        } else {
+                            return [400, "Participant #$p->{seq}, dataset #$h->{dataset}: No argv/args supplied for code"];
+                        }
+                    } else {
+                        $code = $p->{code};
+                    }
+                } elsif (my $template = $p->{code_template} || $p->{fcall_template}) {
+                    my $template_vars;
+                    if ($ds->{args}) {
+                        $template_vars = { %$args };
+                    } elsif ($ds->{argv}) {
+                        $template_vars = { map {$_=>$ds->{argv}[$_]}
+                                               0..$#{ $ds->{argv} } };
+                    } else {
+                        #warn "Item #$item_seq: participant specifies code_template/fcall_template but there is no args/argv in the dataset #$h->{dataset}\n";
+                    }
+
+                    if ($template_vars) {
+                        $template =~ s/\<(\w+(?::\w+)?)\>/dmp($template_vars->{$1})/eg;
+                    }
+                    my $code_str = "sub { $template }";
+                    $log->debugf("Item #%d: code=%s", $item_seq, $code_str);
+                    $code = eval $code_str;
+                    return [400, "Item #$item_seq: code compile error: $@ (code: $code_str)"] if $@;
+                }
+            } else {
+                return [400, "Unknown participant type '$p->{type}'"];
+            }
+
+            # determine item name
+            {
+                # convert participant's & dataset index to name temporarily, for
+                # nicer item name
+                my %h = (%$h_args, %$h);
+                $h{participant} = $p->{name};
+                $h{dataset} = $ds->{name} if $ds;
+                my @k = keys %h;
+                if (@k == 1) {
+                    $item_name = $h{$k[0]};
+                } else {
+                    $item_name = dmp(\%h);
+                }
+                #$log->tracef("Set item name to: %s", $item_name);
+            }
+
+            push @$items, {
+                seq  => $item_seq,
+                name => $item_name,
+                code => $code,
+                _permute => $h,
+                ((_permute_args => $h_args) x !!$ds->{args}),
+            };
+
+            last ITER_ARGS unless $ds->{args};
+        } # ITER_ARGS
+
     } # ITER
 
     $items = _filter_records(
@@ -917,6 +960,11 @@ _
                     summary => 'Shortcut for -a list-datasets',
                     code => sub { $_[0]{action} = 'list-datasets' },
                 },
+                list_permutes => {
+                    is_flag => 1,
+                    summary => 'Shortcut for -a list-permutes',
+                    code => sub { $_[0]{action} = 'list-permutes' },
+                },
                 list_items => {
                     is_flag => 1,
                     summary => 'Shortcut for -a list-items',
@@ -1564,7 +1612,7 @@ Bencher - A benchmark framework
 
 =head1 VERSION
 
-This document describes version 0.18 of Bencher (from Perl distribution Bencher), released on 2015-12-17.
+This document describes version 0.19 of Bencher (from Perl distribution Bencher), released on 2015-12-23.
 
 =head1 SYNOPSIS
 
@@ -1821,7 +1869,7 @@ An example scenario (from C<Bench::Scenario::Example>):
  };
  1;
 
-=head2 Participants
+=head2 participants
 
 B<participants> (array) lists Perl code (or external command) that we want to
 benchmark. Instead of just list of coderefs like what L<Benchmark> expects, you
@@ -1868,7 +1916,7 @@ version will use shell. See L<Bencher::Scenario::Interpreters>.
 
 =back
 
-=head2 Datasets
+=head2 datasets
 
 B<datasets> (array) lists the function inputs (or command-line arguments). You
 can C<name> each dataset too, to be able to refer to it more easily.
@@ -1887,6 +1935,18 @@ benchmark, unless the dataset is explicitly included).
 
 =item * args
 
+Example:
+
+ {filename=>"ujang.txt", size=>10}
+
+You can supply multiple argument values by adding C<@> suffix to the argument
+name, example:
+
+ {filename=>"ujang.txt", 'size@'=>[10, 100, 1000]}
+
+This means, for each participant mentioning C<size>, three benchmark items will
+be generated, one for each value of C<size>.
+
 =item * argv
 
 =item * include_by_default (bool, default 1)
